/*
 * Copyright (C) 2013 Giorgio Vazzana
 *
 * This file is part of Seren.
 *
 * Seren is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Seren is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PC_ENGINE_INTERNAL_H
#define PC_ENGINE_INTERNAL_H

#include "melodies.h"

/* A CALL udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 * Payload:
 *   nick            (NICKLEN bytes)
 */
#define UDP_PACKET_TYPE_CALL    0x0000

/* A CONNECT udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 * Payload:
 *   nick            (NICKLEN bytes)
 */
#define UDP_PACKET_TYPE_CONNECT 0x0001

/* A REFUSE udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_REFUSE  0x0002

/* A TABLE udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   table_size      (uint16, be)
 * Payload:
 *   my_nick         (NICKLEN bytes)
 *  repeat table_size times for all other nodes except caller:
 *   nick            (NICKLEN bytes)
 *   addr            (uint32, be)
 *   port            (uint16, be)
 */
#define UDP_PACKET_TYPE_TABLE   0x0003

/* A BYE udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_BYE     0x0004


/* An AUDIO udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   sequence_number (uint32, be)
 * Payload:
 *   data            (variable size, 4000 byte max)
 *
 * The field sequence_number must start at 0 for the first packet and increase
 * monotonically. The payload data is the opus packet coming out of the encoder.
 */
#define UDP_PACKET_TYPE_AUDIO   0x0100

/* A NOP udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 */
#define UDP_PACKET_TYPE_NOP     0x0101

/* A CHAT udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 * Payload:
 *   data            (variable size)
 */
#define UDP_PACKET_TYPE_CHAT    0x0102


/* A RELAY udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   dest_addr       (uint32, be)
 *   dest_port       (uint16, be)
 * Payload:
 *   packet to relay (variable size)
 */
#define UDP_PACKET_TYPE_RELAY   0x0200

/* A RELAYED udp packet has the following structure:
 * Header:
 *   packet_type     (uint16, be)
 *   source_addr     (uint32, be)
 *   source_port     (uint16, be)
 * Payload:
 *   packet relayed (variable size)
 */
#define UDP_PACKET_TYPE_RELAYED 0x0201


/* A ENCRYPTED udp packet has the following structure:
 * Header:
 *   packet_type           (uint16, be)
 *   initialization_vector (uint64, be)
 * Payload:
 *   encrypted packet      (variable size)
 *
 * The payload is encrypted with xtea and must have a size multiple of 8 bytes.
 * To obtain this, a padding at the end of the packet is added, which consists
 * of exactly n bytes all with value n. If the size of the packet is already
 * multiple of 8, another 8 bytes will be added, all with value 8.
 * This means that n>0 always and (packetlen + n) % 8 = 0
 */
#define UDP_PACKET_TYPE_ENCRYPTED 0x8000


#define UDP_PACKET_AUDIO_HEADER_LEN  6
#define UDP_PACKET_AUDIO_PAYLOAD_LEN (4000)
//#define UDP_PACKET_MAXLEN            (UDP_PACKET_AUDIO_HEADER_LEN + UDP_PACKET_AUDIO_PAYLOAD_LEN + 2 + 8 + 8)

/* node status */
#define STATUS_CONNECTING     0
#define STATUS_CONNECTING_END 99
#define STATUS_CALLING        100
#define STATUS_OK             200

/* reasons for deleting a node */
#define REASON_LEFT         0
#define REASON_TIMEOUT      1

static int is_same_addr(struct sockaddr_in a, struct sockaddr_in b);

static struct pc_event *event_push(struct slist **events, enum pc_event_type type);

//static void sendrecipe_delete(struct slist **sendrecipes, struct sockaddr_in addr);

static struct pc_caller *caller_push(struct slist **callers, struct sockaddr_in addr, const char *nick);
static struct pc_caller *caller_is_present(struct slist *callers, struct sockaddr_in addr);

static unsigned int node_get_count(struct pc_node *nodes[]);
static int  node_get_idx(struct pc_node *nodes[], struct sockaddr_in addr);
static void node_delete(struct pc_node *nodes[], int idx, int reason, struct pc_tone *tone, struct slist **events);
static int  node_add(struct pc_node *nodes[], struct sockaddr_in addr, const char *nick, struct pc_tone *tone, struct slist **events);

static void packet_call(struct udp_packet *udp_packet, const char *nick);
static void packet_connect(struct udp_packet *udp_packet, const char *nick);
static void packet_refuse(struct udp_packet *udp_packet);
static void packet_table(struct udp_packet *udp_packet, const char *nick, struct pc_node *nodes[], int exclude_idx);
static void packet_bye(struct udp_packet *udp_packet);
static void packet_audio(struct udp_packet *udp_packet, uint32_t sequence_number, uint8_t *payload, size_t payloadlen);
static void packet_nop(struct udp_packet *udp_packet);
static void packet_chat(struct udp_packet *udp_packet, const char *buf, size_t len);
static void packet_relay(struct udp_packet *udp_packet, struct sockaddr_in dest_addr);
static void packet_from_relay_to_relayed(struct udp_packet *udp_packet);
static void packet_encrypt(struct udp_packet *udp_packet, const uint32_t key[4]);

static void send_udp_packet_enc(struct pc_context *pctx, const struct udp_packet *udp_packet);
static void send_udp_packet_enc_to_all(struct pc_context *pctx, const struct udp_packet *udp_packet);

static void connect_new_nodes(struct pc_context *pctx, int use_relay);

static int  process_packet(struct pc_context *pctx, const struct udp_packet *udp_packet);

static opus_int32 convert_opus_bw(opus_int32 bandwidth);
static void update_bitrate_and_bandwidth(struct pc_context *pctx);
static double sinc(double x);
static void generate_tone(struct pc_tone *tone);
static void generate_calltone(struct pc_calltone *calltone);
static void generate_ringtone(struct pc_ringtone *ringtone);
static int32_t find_peak_s16_2ch(const int16_t *pcm, size_t samples);
static int32_t find_peak_s32_2ch(const int32_t *pcm, size_t samples);
static uint64_t sum_of_squares(const int16_t *pcm, size_t samples);
static unsigned int mix_audio(struct pc_context *pctx);
static uint8_t hex2num(uint8_t c);

#endif
