/*
 * Copyright (C) 2013 Giorgio Vazzana
 *
 * This file is part of Seren.
 *
 * Seren is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Seren is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <poll.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "pc-engine.h"
#include "pc-engine-internal.h"
#include "common.h"
#include "rw.h"
#include "msgbook.h"
#include "xtea.h"
#include "recording.h"

#define MODULE "pc-engine"

const unsigned int samplerate = 48000, channels = 2;
unsigned int fd_idx_stdin;
unsigned int fd_idx_udpsocket;
unsigned int fd_idx_audiocapture;
unsigned int fd_idx_audioplayback;

static int is_same_addr(struct sockaddr_in a, struct sockaddr_in b)
{
	if (a.sin_addr.s_addr == b.sin_addr.s_addr && a.sin_port == b.sin_port)
		return 1;
	else
		return 0;
}


static struct pc_event *event_push(struct slist **events, enum pc_event_type type)
{
	struct pc_event *ev;

	ev = xcalloc(1, sizeof(*ev));
	ev->type = type;
	*events = slist_append(*events, ev);

	return ev;
}

#if 0
static void sendrecipe_delete(struct slist **sendrecipes, struct sockaddr_in addr)
{
	struct slist *p;

	p = *sendrecipes;
	while (p) {
		struct slist *pnext = p->next;
		struct pc_sendrecipe *sr = p->data;

		if (is_same_addr(addr, sr->addr)) {
			free(sr);
			*sendrecipes = slist_remove_element(*sendrecipes, p);
		}
		p = pnext;
	}
}
#endif

static struct pc_caller *caller_push(struct slist **callers, struct sockaddr_in addr, const char *nick)
{
	struct pc_caller *cl;

	cl = xcalloc(1, sizeof(*cl));
	cl->addr = addr;
	memcpy(cl->nick, nick, NICKLEN);
	cl->timeout = 250; /* 5s */
	*callers = slist_append(*callers, cl);

	return cl;
}

static struct pc_caller *caller_is_present(struct slist *callers, struct sockaddr_in addr)
{
	struct slist *p;

	p = callers;
	while (p) {
		struct pc_caller *cl = p->data;

		if (is_same_addr(addr, cl->addr))
			return cl;

		p = p->next;
	}
	return NULL;
}


static unsigned int node_get_count(struct pc_node *nodes[])
{
	unsigned int i, n = 0;

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i])
			n++;
	}
	return n;
}

static int node_get_idx(struct pc_node *nodes[], struct sockaddr_in addr)
{
	int i;

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] && is_same_addr(addr, nodes[i]->addr))
			return i;
	}
	return -1;
}

static void node_delete(struct pc_node *nodes[], int idx, int reason, struct pc_tone *tone, struct slist **events)
{
	struct pc_event *ev;
	struct slist *p;

	/* push event */
	ev = event_push(events, event_type_node_delete);
	ev->addr = nodes[idx]->addr;
	memcpy(ev->nick, nodes[idx]->nick, NICKLEN);
	ev->i = (reason == REASON_TIMEOUT);

	tone->tone_play = (tone->tone_play > 10*ONE_OPUS_FRAME) ? tone->tone_play : 10*ONE_OPUS_FRAME;

	/* destroy decoder state */
	opus_decoder_destroy(nodes[idx]->dec);

	/* free pcm opus frames */
	p = nodes[idx]->pcmframes;
	while (p) {
		struct slist *pnext = p->next;
		struct pc_pcmframe *pf = p->data;

		free(pf);
		nodes[idx]->pcmframes = slist_remove_element(nodes[idx]->pcmframes, p);
		p = pnext;
	}

	free(nodes[idx]);
	nodes[idx] = NULL;
}

static int node_add(struct pc_node *nodes[], struct sockaddr_in addr, const char *nick, struct pc_tone *tone, struct slist **events)
{
	int i, idx, error;
	struct pc_event *ev;

	/* check if we have space for another node */
	idx = -1;
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] == NULL) {
			idx = i;
			break;
		}
	}
	if (idx == -1) {
		snprintf(msgbuf, MBS, "Max node capacity reached (%d nodes)", MAX_NUMBER_OF_NODES);
		msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
		return -1;
	}

	/* allocate and clear memory for new node */
	nodes[idx] = xcalloc(1, sizeof(struct pc_node));

	/* set non-zero fields */
	nodes[idx]->addr                 = addr;
	if (nick != NULL)
		memcpy(nodes[idx]->nick, nick, NICKLEN);
	else
		memset(nodes[idx]->nick,    0, NICKLEN);
	nodes[idx]->timeout              = TIMEOUT;

	/* create decoder state */
	nodes[idx]->dec = opus_decoder_create(samplerate, channels, &error);
	if (error != OPUS_OK) {
		free(nodes[idx]);
		nodes[idx] = NULL;
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
		return -1;
	}

	tone->tone_play = (tone->tone_play > 20*ONE_OPUS_FRAME) ? tone->tone_play : 20*ONE_OPUS_FRAME;

	/* push event */
	ev = event_push(events, event_type_node_add);
	ev->addr = nodes[idx]->addr;
	memcpy(ev->nick, nodes[idx]->nick, NICKLEN);

	return idx;
}


static void packet_call(struct udp_packet *udp_packet, const char *nick)
{
	udp_packet->len = 2+NICKLEN;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CALL);
	memcpy(udp_packet->data+2, nick, NICKLEN);
}

static void packet_connect(struct udp_packet *udp_packet, const char *nick)
{
	udp_packet->len = 2+NICKLEN;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CONNECT);
	memcpy(udp_packet->data+2, nick, NICKLEN);
}

static void packet_refuse(struct udp_packet *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_REFUSE);
}

static void packet_table(struct udp_packet *udp_packet, const char *nick, struct pc_node *nodes[], int exclude_idx)
{
	int i;
	uint8_t *d;
	size_t   l;
	uint16_t table_size;

	d = udp_packet->data;
	l = 0;

	/* all nodes excluding the sender of the call request */
	table_size = node_get_count(nodes);

	/* nodes address table */
	write_be16(d, UDP_PACKET_TYPE_TABLE); d+=2; l+=2;
	/* table size */
	write_be16(d, table_size);            d+=2; l+=2;
	/* my nick */
	memcpy(d, nick, NICKLEN);             d+=NICKLEN; l+=NICKLEN;
	/* other nodes except packet sender */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (nodes[i] == NULL)
			continue;
		if (i == exclude_idx)
			continue;
		memcpy(d, nodes[i]->nick, NICKLEN);                   d+=NICKLEN; l+=NICKLEN;
		write_be32(d, ntohl(nodes[i]->addr.sin_addr.s_addr)); d+=4; l+=4;
		write_be16(d, ntohs(nodes[i]->addr.sin_port));        d+=2; l+=2;
	}
	udp_packet->len = l;
}

static void packet_bye(struct udp_packet *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_BYE);
}

static void packet_audio(struct udp_packet *udp_packet, uint32_t sequence_number, uint8_t *payload, size_t payloadlen)
{
	udp_packet->len = UDP_PACKET_AUDIO_HEADER_LEN + payloadlen;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_AUDIO);
	write_be32(udp_packet->data+2, sequence_number);
	memcpy(udp_packet->data + UDP_PACKET_AUDIO_HEADER_LEN, payload, payloadlen);
}

static void packet_nop(struct udp_packet *udp_packet)
{
	udp_packet->len = 2;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_NOP);
}

static void packet_chat(struct udp_packet *udp_packet, const char *buf, size_t len)
{
	udp_packet->len = 2+len;
	write_be16(udp_packet->data, UDP_PACKET_TYPE_CHAT);
	memcpy(udp_packet->data+2, buf, len);
}

static void packet_relay(struct udp_packet *udp_packet, struct sockaddr_in dest_addr)
{
	uint8_t *d;
	size_t   l;

	d = udp_packet->data;
	l = 0;
	memmove(d+8, d, udp_packet->len);

	write_be16(d, UDP_PACKET_TYPE_RELAY);            d+=2; l+=2;
	write_be32(d, ntohl(dest_addr.sin_addr.s_addr)); d+=4; l+=4;
	write_be16(d, ntohs(dest_addr.sin_port));        d+=2; l+=2;
	udp_packet->len += l;
}

static void packet_from_relay_to_relayed(struct udp_packet *udp_packet)
{
	uint8_t *d;
	size_t   l;
/*	uint16_t packet_type;*/
	struct sockaddr_in src_addr, dest_addr;

	/* save source address */
	src_addr = udp_packet->addr;

	/* read destination address */
	memset(&dest_addr, 0, sizeof(dest_addr));
	d = udp_packet->data;
	l = udp_packet->len;
/*	packet_type = read_be16(d);*/                    d+=2; l-=2;
	dest_addr.sin_family      = AF_INET;
	dest_addr.sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
	dest_addr.sin_port        = htons(read_be16(d)); d+=2; l-=2;

	/* this now becomes a relayed packet */
	udp_packet->addr = dest_addr;
	d = udp_packet->data;
	l = 0;
	write_be16(d, UDP_PACKET_TYPE_RELAYED);          d+=2; l+=2;
	write_be32(d, ntohl(src_addr.sin_addr.s_addr));  d+=4; l+=4;
	write_be16(d, ntohs(src_addr.sin_port));         d+=2; l+=2;
}

static void packet_encrypt(struct udp_packet *udp_packet, const uint32_t key[4])
{
	uint8_t *d;
	size_t   l;
	uint32_t iv[2];
	uint8_t  padding_size;

	/* insert some padding to make payload size multiple of 8 */
	padding_size = 8 - (udp_packet->len % 8);
	memset(udp_packet->data + udp_packet->len, padding_size, padding_size);
	udp_packet->len += padding_size;

	/* encrypt payload in packet */
	iv[0] = rand();
	iv[1] = rand();
	xtea_encrypt_buffer_cbc(udp_packet->data, udp_packet->len, key, iv);

	/* insert header */
	d = udp_packet->data;
	l = 0;
	memmove(d+10, d, udp_packet->len);

	write_be16(d, UDP_PACKET_TYPE_ENCRYPTED); d+=2; l+=2;
	write_be32(d, iv[0]);                     d+=4; l+=4;
	write_be32(d, iv[1]);                     d+=4; l+=4;
	udp_packet->len += l;
}


static void send_udp_packet_enc(struct pc_context *pctx, const struct udp_packet *udp_packet)
{
	if (pctx->encrypt) {
		struct udp_packet udp_packet_enc = *udp_packet;

		packet_encrypt(&udp_packet_enc, pctx->key);
		send_udp_packet(pctx->sockfd, &udp_packet_enc);
		pctx->bytes_out_per_sec += udp_packet_enc.len;
	} else {
		send_udp_packet(pctx->sockfd, udp_packet);
		pctx->bytes_out_per_sec += udp_packet->len;
	}
}

static void send_udp_packet_enc_to_all(struct pc_context *pctx, const struct udp_packet *udp_packet)
{
	unsigned int i;
	struct udp_packet udp_packet_enc = *udp_packet;

	if (pctx->encrypt)
		packet_encrypt(&udp_packet_enc, pctx->key);

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i] && pctx->nodes[i]->status == STATUS_OK) {
			udp_packet_enc.addr = pctx->nodes[i]->addr;
			send_udp_packet(pctx->sockfd, &udp_packet_enc);
			pctx->bytes_out_per_sec += udp_packet_enc.len;
		}
	}
}


static void connect_new_nodes(struct pc_context *pctx, int use_relay)
{
	unsigned int i, j, relay_count;
	struct udp_packet udp_packet;

	/* send connect requests to those who don't know about me */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i] && pctx->nodes[i]->status < STATUS_CONNECTING_END) {
			udp_packet.addr = pctx->nodes[i]->addr;
			packet_connect(&udp_packet, pctx->nick);
			send_udp_packet_enc(pctx, &udp_packet);
			pctx->nodes[i]->status++;

			if (use_relay || pctx->nodes[i]->status >= (STATUS_CONNECTING+3)) {
				/* prepare packet for relay */
				packet_relay(&udp_packet, pctx->nodes[i]->addr);

				/* send this packet to (at most) 2 other nodes that can relay it */
				relay_count = 2;
				for (j = 0; j < MAX_NUMBER_OF_NODES; j++) {
					if (pctx->nodes[j] && pctx->nodes[j]->status == STATUS_OK && j != i) {
						udp_packet.addr = pctx->nodes[j]->addr;
						send_udp_packet_enc(pctx, &udp_packet);
						relay_count--;
						if (relay_count == 0)
							break;
					}
				}

				if (relay_count < 2) {
					snprintf(msgbuf, MBS, "Cannot connect to %s:%hu (%s), sent %d relay packets",
							 inet_ntoa(pctx->nodes[i]->addr.sin_addr), ntohs(pctx->nodes[i]->addr.sin_port),
							 pctx->nodes[i]->nick, 2-relay_count);
					msgbook_enqueue(&mb0, MB_TYPE_WARNING, MODULE, msgbuf);
				}
			}
		}
	}
}


struct address_book {
	int                n;
	char               nick[MAX_NUMBER_OF_NODES][NICKLEN];
	struct sockaddr_in addr[MAX_NUMBER_OF_NODES];
};

static int process_packet(struct pc_context *pctx, const struct udp_packet *udp_packet)
{
	int i, j, idx;
	struct pc_node **nodes  = pctx->nodes;
	struct pc_tone  *tone   = &pctx->tone;
	struct slist   **events = &pctx->events;
	const uint8_t   *d;
	size_t           l;
	uint16_t packet_type, table_size;
	struct address_book ab;
	struct udp_packet udp_packet_response;
	struct pc_event *ev;

	if (udp_packet->len < 2)
		return RES_ERROR;

	d = udp_packet->data;
	l = udp_packet->len;
	packet_type = read_be16(d); d+=2; l-=2;

	/* do we know this node already? */
	idx = node_get_idx(nodes, udp_packet->addr);
	if (idx == -1 &&
	    packet_type != UDP_PACKET_TYPE_CALL &&
	    packet_type != UDP_PACKET_TYPE_CONNECT) {
		if (pctx->verbose >= 2) {
			snprintf(msgbuf, MBS, "Unexpected 0x%04x packet from: %s:%hu", packet_type,
				     inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
		}
		return RES_ERROR;
	}

	/* reset timeout counter */
	if (idx != -1)
		nodes[idx]->timeout = TIMEOUT;


	switch (packet_type) {
	case UDP_PACKET_TYPE_CALL:
		if (l != NICKLEN) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid CALL packet received");
			return RES_ERROR;
		}

		if (idx != -1) {
			/* node is already present, send table */
			udp_packet_response.addr = udp_packet->addr;
			packet_table(&udp_packet_response, pctx->nick, nodes, idx);
			send_udp_packet_enc(pctx, &udp_packet_response);
		} else {
			struct pc_caller *cl;

			cl = caller_is_present(pctx->callers, udp_packet->addr);
			if (cl) {
				cl->timeout = 250;
			} else {
				/* push caller */
				caller_push(&pctx->callers, udp_packet->addr, (char *)d);
				/* push event */
				ev = event_push(events, event_type_incoming_call);
				ev->addr = udp_packet->addr;
				memcpy(ev->nick, (char *)d, NICKLEN);

				if (pctx->autoaccept_calls)
					pc_engine_cmd_accept_call(pctx);
			}
		}
		return RES_OK;


	case UDP_PACKET_TYPE_CONNECT:
		if (l != NICKLEN) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid CONNECT packet received");
			return RES_ERROR;
		}

		/* insert remote node */
		if (idx == -1) {
			idx = node_add(nodes, udp_packet->addr, (char *)d, tone, events);
			if (idx == -1) {
				/* send refuse */
				udp_packet_response.addr = udp_packet->addr;
				packet_refuse(&udp_packet_response);
				send_udp_packet_enc(pctx, &udp_packet_response);
				return RES_REFUSED;
			}
			nodes[idx]->status = STATUS_OK;
		}

		/* send table */
		udp_packet_response.addr = udp_packet->addr;
		packet_table(&udp_packet_response, pctx->nick, nodes, idx);
		send_udp_packet_enc(pctx, &udp_packet_response);
		return RES_OK;


	case UDP_PACKET_TYPE_REFUSE:
		if (l != 0) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid REFUSE packet received");
			return RES_ERROR;
		}
		/* push event */
		ev = event_push(events, event_type_call_refused);
		ev->addr = nodes[idx]->addr;
		memcpy(ev->nick, nodes[idx]->nick, NICKLEN);

		/* stop calling */
		if (pctx->call && is_same_addr(pctx->call->addr, nodes[idx]->addr)) {
			free(pctx->call);
			pctx->call = NULL;
		}

		node_delete(nodes, idx, REASON_LEFT, tone, events);
		return RES_REFUSED;


	case UDP_PACKET_TYPE_TABLE:
		table_size = read_be16(d); d+=2; l-=2;
		if (l != (size_t)(NICKLEN*table_size + 6*(table_size-1))) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid TABLE packet received");
			return RES_ERROR;
		}

		/* stop calling */
		if (pctx->call && is_same_addr(pctx->call->addr, nodes[idx]->addr)) {
			free(pctx->call);
			pctx->call = NULL;
		}

		/* copy table to addres book */
		ab.n = table_size;
		for (i = 0; i < ab.n; i++) {
			memcpy(ab.nick[i], d, NICKLEN);                       d+=NICKLEN; l-=NICKLEN;
			if (i == 0) {
				ab.addr[i] = udp_packet->addr;
			} else {
				ab.addr[i].sin_family      = AF_INET;
				ab.addr[i].sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
				ab.addr[i].sin_port        = htons(read_be16(d)); d+=2; l-=2;
			}
		}

		/* copy sender's nick */
		memcpy(nodes[idx]->nick, ab.nick[0], NICKLEN);

		/* packet sender knows about me */
		nodes[idx]->status = STATUS_OK;

		/* push event */
		ev = event_push(events, event_type_call_accepted);
		ev->addr = nodes[idx]->addr;
		memcpy(ev->nick, nodes[idx]->nick, NICKLEN);

		/* insert unknown nodes */
		for (j = 0; j < ab.n; j++) {
			int tidx;

			tidx = node_get_idx(nodes, ab.addr[j]);
			if (tidx == -1)
				node_add(nodes, ab.addr[j], ab.nick[j], tone, events);
		}

		/* send connect requests to newly added nodes. we don't use relay initially */
		//connect_new_nodes(pctx, 0);
		return RES_OK;


	case UDP_PACKET_TYPE_BYE:
		if (l != 0) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid BYE packet received");
			return RES_ERROR;
		}
		node_delete(nodes, idx, REASON_LEFT, tone, events);
		return RES_OK;


	case UDP_PACKET_TYPE_AUDIO:
		/* read packet header */
		nodes[idx]->sequence_number = read_be32(d); d+=4; l-=4;

		/* ok, sequence numbers of packets must always increase */
		if (nodes[idx]->sequence_number > nodes[idx]->last_sequence_number || nodes[idx]->packets_received == 0) {
			struct pc_pcmframe *pf;
			const uint8_t *opus_dec_packet;
			opus_int32     opus_dec_packetlen;
			opus_int32     pcmlen;

			/* check if we lost packets */
			if (nodes[idx]->sequence_number != nodes[idx]->last_sequence_number + 1 &&
			    nodes[idx]->packets_received != 0) {
				uint32_t k, loss;

				loss = nodes[idx]->sequence_number - (nodes[idx]->last_sequence_number + 1);
				nodes[idx]->packets_lost += loss;
				if (pctx->verbose >= 2) {
					snprintf(msgbuf, MBS, "Oops, %u audio packets lost", loss);
					msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
				}

				/* call decoder for missing packets if the loss is less than 25 packets */
				for (k = 0; loss < 25 && k < loss; k++) {
					pf = xcalloc(1, sizeof(*pf));
					pf->pcmlen = ONE_OPUS_FRAME;
					pcmlen = opus_decode(nodes[idx]->dec, NULL, 0, pf->pcm, pf->pcmlen, 0);

					if (pcmlen != ONE_OPUS_FRAME || nodes[idx]->nb_frames >= 25) {
						free(pf);
					} else {
						/* add one opus frame to the tail of the queue */
						nodes[idx]->pcmframes = slist_append(nodes[idx]->pcmframes, pf);
						nodes[idx]->nb_frames++;
					}
				}
			}

			nodes[idx]->last_sequence_number = nodes[idx]->sequence_number;
			nodes[idx]->packets_received++;

			/* decode */
			pf = xcalloc(1, sizeof(*pf));
			pf->pcmlen = ONE_OPUS_FRAME;
			opus_dec_packet    = udp_packet->data + UDP_PACKET_AUDIO_HEADER_LEN;
			opus_dec_packetlen = udp_packet->len  - UDP_PACKET_AUDIO_HEADER_LEN;
			pcmlen = opus_decode(nodes[idx]->dec, opus_dec_packet, opus_dec_packetlen, pf->pcm, pf->pcmlen, 0);

			if (pcmlen != ONE_OPUS_FRAME) {
				free(pf);
				if (pcmlen < 0)
					snprintf(msgbuf, MBS, "opus_decode() failed: %s", opus_strerror(pcmlen));
				else
					snprintf(msgbuf, MBS, "pcmlen = %d, discarding pcm data", pcmlen);
				msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
				return RES_ERROR;
			}

			/* limit queue length to 500ms */
			if (nodes[idx]->nb_frames >= 25) {
				free(pf);
			} else {
				/* add one opus frame to the tail of the queue */
				nodes[idx]->pcmframes = slist_append(nodes[idx]->pcmframes, pf);
				nodes[idx]->nb_frames++;
			}
		} else {
			if (pctx->verbose >= 2)
				msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, "Oops, audio packet duplicated or out of order");
		}
		return RES_OK;


	case UDP_PACKET_TYPE_NOP:
		if (l != 0) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid NOP packet received");
			return RES_ERROR;
		}
		return RES_OK;


	case UDP_PACKET_TYPE_CHAT:
		/* push event */
		ev = event_push(events, event_type_chat);
		ev->addr = nodes[idx]->addr;
		memcpy(ev->nick, nodes[idx]->nick, NICKLEN);
		ev->data = xmalloc(l);
		memcpy(ev->data, d, l);

		tone->tone_play = (tone->tone_play > 5*ONE_OPUS_FRAME) ? tone->tone_play : 5*ONE_OPUS_FRAME;
		return RES_OK;


	case UDP_PACKET_TYPE_RELAY:
		if (l < 6) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid RELAY packet received");
			return RES_ERROR;
		}

		/* copy received packet */
		udp_packet_response = *udp_packet;

		/* this now becomes a relayed packet */
		packet_from_relay_to_relayed(&udp_packet_response);

		if (pctx->verbose >= 2) {
			/* use 2 calls because inet_ntoa returns a string in a statically allocated buffer,
			   which subsequent calls will overwrite */
			snprintf(msgbuf, MBS, "Relay packet from %s:%hu",
				     inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
			snprintf(msgbuf, MBS, ", relaying to %s:%hu",
				     inet_ntoa(udp_packet_response.addr.sin_addr), ntohs(udp_packet_response.addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
		}

		/* send relayed packet */
		send_udp_packet_enc(pctx, &udp_packet_response);
		return RES_OK;


	case UDP_PACKET_TYPE_RELAYED:
		if (l < 6) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid RELAYED packet received");
			return RES_ERROR;
		}

		/* copy received packet, discarding relayed header */
		memcpy(udp_packet_response.data, udp_packet->data+8, udp_packet->len-8);
		udp_packet_response.len = udp_packet->len-8;

		/* extract original source */
		udp_packet_response.addr.sin_family      = AF_INET;
		udp_packet_response.addr.sin_addr.s_addr = htonl(read_be32(d)); d+=4; l-=4;
		udp_packet_response.addr.sin_port        = htons(read_be16(d)); d+=2; l-=2;

		if (pctx->verbose >= 2) {
			/* use 2 calls because inet_ntoa returns a string in a statically allocated buffer,
			   which subsequent calls will overwrite */
			snprintf(msgbuf, MBS, "Relayed packet from %s:%hu",
				     inet_ntoa(udp_packet_response.addr.sin_addr), ntohs(udp_packet_response.addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
			snprintf(msgbuf, MBS, " via %s:%hu",
				     inet_ntoa(udp_packet->addr.sin_addr), ntohs(udp_packet->addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
		}

		return process_packet(pctx, &udp_packet_response);


	default:
		if (pctx->verbose >= 2)
			msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, "Unknown packet type");
		break;
	}
	return RES_ERROR;
}


static opus_int32 convert_opus_bw(opus_int32 bandwidth)
{
	switch (bandwidth) {
	case OPUS_BANDWIDTH_NARROWBAND:
		return 4000;
	case OPUS_BANDWIDTH_MEDIUMBAND:
		return 6000;
	case OPUS_BANDWIDTH_WIDEBAND:
		return 8000;
	case OPUS_BANDWIDTH_SUPERWIDEBAND:
		return 12000;
	case OPUS_BANDWIDTH_FULLBAND:
		return 20000;
	default: /* AUTO */
		return 0;
	}
}

static void update_bitrate_and_bandwidth(struct pc_context *pctx)
{
	int error;

	/* get bitrate */
	error = opus_encoder_ctl(pctx->enc, OPUS_GET_BITRATE(&pctx->bitrate));
	if (error != OPUS_OK)
		pctx->bitrate = 0;

	/* get bandwidth */
	error = opus_encoder_ctl(pctx->enc, OPUS_GET_BANDWIDTH(&pctx->bandwidth));
	if (error != OPUS_OK)
		pctx->bandwidth_hz = 0;
	else
		pctx->bandwidth_hz = convert_opus_bw(pctx->bandwidth);
}

static double sinc(double x)
{
	if (x == 0.0)
		return 1.0;

	return (sin(M_PI*x) / (M_PI*x));
}

static void generate_tone(struct pc_tone *tone)
{
	unsigned int i;
	double v;

	for (i = 0; i < TONEBUFLEN; i++) {
		v = 10000.0 * sin(2.0 * M_PI * 440.0 * i / (double)samplerate);
		if (channels == 1) {
			tone->pcm[i] = lrint(v);
		} else {
			tone->pcm[channels*i    ] = lrint(v);
			tone->pcm[channels*i + 1] = lrint(v);
		}
	}
	tone->tone_pos  = 0;
	tone->tone_play = 0;
}

static void generate_calltone(struct pc_calltone *calltone)
{
	unsigned int i;
	double v;

	/* ITU-T E.180 ringing tone for Italy, 425Hz, 4/1, start at 0.5s */
	for (i = 25*ONE_OPUS_FRAME; i < 75*ONE_OPUS_FRAME; i++) {

		v = 20000.0 * sin(2.0 * M_PI * 425.0 * i / (double)samplerate);
		if (channels == 1) {
			calltone->pcm[i] = lrint(v);
		} else {
			calltone->pcm[channels*i    ] = lrint(v);
			calltone->pcm[channels*i + 1] = lrint(v);
		}
	}
	calltone->pos = 0;
}

static void generate_ringtone(struct pc_ringtone *ringtone)
{
	unsigned int n, i, samples, pos;
	double duration, v;
	const struct note *song = SONG;

	/* compute ringtone duration */
	duration = 0.0;
	for (n = 0; song[n].duration != 0; n++)
		duration += song[n].duration;
	duration /= 1000.0;
	duration += 0.5 + 1.5; /* start silence + end silence */

	/* allocate memory for the samples buffer */
	ringtone->pcmlen = (((unsigned int)(duration * samplerate)) / ONE_OPUS_FRAME) * ONE_OPUS_FRAME;
	ringtone->pcm    = xcalloc(channels * sizeof(*ringtone->pcm), ringtone->pcmlen);

	/* synthesize ringtone */
	pos = channels * 25*ONE_OPUS_FRAME; /* skip 0.5s */
	for (n = 0; song[n].duration != 0; n++) {
		samples = samplerate * song[n].duration / 1000;
		for (i = 0; i < samples; i++) {
			v = 20000.0 * sin(2.0 * M_PI * song[n].freq * i / (double)samplerate);
			/* square wave */
			//v = v < 0.0 ? -1 : 1;
			/* exponential modulation */
			//v *= exp(-(double)i / (1.0 * samples));
			/* sinc modulation */
			v *= sinc(-(double)i * 2.0 / (double)samplerate);

			if (channels == 1) {
				ringtone->pcm[i + pos] = lrint(v);
			} else {
				ringtone->pcm[channels*i + pos    ] = lrint(v);
				ringtone->pcm[channels*i + pos + 1] = lrint(v);
			}
		}
		pos += channels*samples;
	}

/*
	for (i = 0; i < RINGTONEBUFLEN; i++) {
		v = 20000.0 * sin(2.0 * M_PI * 2.0 * i / (double)samplerate) * sin(2.0 * M_PI * 440.0 * i / (double)samplerate);
		if (channels == 1) {
			ringtone->pcm[i] = lrint(v);
		} else {
			ringtone->pcm[channels*i    ] = lrint(v);
			ringtone->pcm[channels*i + 1] = lrint(v);
		}
	}
*/
	ringtone->pos = 0;
}

static int32_t find_peak_s16_2ch(const int16_t *pcm, size_t samples)
{
	size_t i;
	int32_t vl, vr, pl, pr, nl, nr;

	pl = pr = nl = nr = 0;
	for (i = 0; i < samples; i += 2) {
		vl = pcm[i  ];
		vr = pcm[i+1];
		if (vl > pl)
			pl = vl;
		if (vl < nl)
			nl = vl;
		if (vr > pr)
			pr = vr;
		if (vr < nr)
			nr = vr;
	}
	pl = pl > -nl ? pl : -nl;
	pr = pr > -nr ? pr : -nr;
	return pl > pr ? pl : pr;
}

static int32_t find_peak_s32_2ch(const int32_t *pcm, size_t samples)
{
	size_t i;
	int32_t vl, vr, pl, pr, nl, nr;

	pl = pr = nl = nr = 0;
	for (i = 0; i < samples; i += 2) {
		vl = pcm[i  ];
		vr = pcm[i+1];
		if (vl > pl)
			pl = vl;
		if (vl < nl)
			nl = vl;
		if (vr > pr)
			pr = vr;
		if (vr < nr)
			nr = vr;
	}
	pl = pl > -nl ? pl : -nl;
	pr = pr > -nr ? pr : -nr;
	return pl > pr ? pl : pr;
}

static uint64_t sum_of_squares(const int16_t *pcm, size_t samples)
{
	size_t i;
	int64_t vl, vr;
	uint64_t suml, sumr;

	suml = sumr = 0;
	for (i = 0; i < samples; i += 2) {
		vl = pcm[i  ];
		vr = pcm[i+1];
		vl *= vl;
		vr *= vr;
		suml += vl;
		sumr += vr;
	}
	return suml + sumr;
}

static unsigned int mix_audio(struct pc_context *pctx)
{
	unsigned int i, j, n, m;
	opus_int32 pcm32[2*ONE_OPUS_FRAME], peak;
	float gain;

	memset(pcm32, 0, sizeof(pcm32));

	n = 0;
	if (pctx->tone.tone_pos >= pctx->tone.tone_play) {
		pctx->tone.tone_pos  = 0;
		pctx->tone.tone_play = 0;
	} else {
		n++;

		/* sum (mix) pcm from tone buffer */
		for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
			pcm32[j] += pctx->tone.pcm[channels * pctx->tone.tone_pos + j];

		pctx->tone.tone_pos += ONE_OPUS_FRAME;
	}

	if (pctx->call) {
		n++;

		if (pctx->calltone.pos >= CALLTONEBUFLEN)
			pctx->calltone.pos = 0;

		/* sum (mix) pcm from ringtone buffer */
		for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
			pcm32[j] += pctx->calltone.pcm[channels * pctx->calltone.pos + j];

		pctx->calltone.pos += ONE_OPUS_FRAME;
	} else {
		pctx->calltone.pos = 0;
	}

	if (pctx->callers) {
		n++;

		if (pctx->ringtone.pos >= pctx->ringtone.pcmlen)
			pctx->ringtone.pos = 0;

		/* sum (mix) pcm from ringtone buffer */
		for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
			pcm32[j] += pctx->ringtone.pcm[channels * pctx->ringtone.pos + j];

		pctx->ringtone.pos += ONE_OPUS_FRAME;
	} else {
		pctx->ringtone.pos = 0;
	}

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i] && pctx->nodes[i]->pcmframes) {
			struct pc_pcmframe *pf;

			n++;
			pf = pctx->nodes[i]->pcmframes->data;

			/* sum (mix) pcm from different nodes */
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pcm32[j] += pf->pcm[j];

			/* remove one opus frame from the head of the queue */
			free(pf);
			pctx->nodes[i]->pcmframes = slist_remove_head(pctx->nodes[i]->pcmframes);
			pctx->nodes[i]->nb_frames--;
		}
	}

	m = 0;
	if (pctx->lb) {
		m++;

		/* sum (mix) pcm from loopback buffer */
		for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
			pcm32[j] += pctx->lb_pcm[j];
	}

	if (n+m) {
		peak = find_peak_s32_2ch(pcm32, channels * ONE_OPUS_FRAME);

		gain = (float)(INT16_MAX) / (float)peak;
		gain = (gain > 1.0f) ? 1.0f : gain;

		if (gain != 1.0f) {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->pcm[j] = lrintf((float)pcm32[j] * gain);
		} else {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->pcm[j] = pcm32[j];
		}

		pctx->pcmlen = ONE_OPUS_FRAME;
	} else {
		pctx->pcmlen = 0;
	}

	if (pctx->record) {
		if (!m) {
			/* sum (mix) pcm from loopback buffer */
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pcm32[j] += pctx->lb_pcm[j];
		}

		peak = find_peak_s32_2ch(pcm32, channels * ONE_OPUS_FRAME);

		gain = (float)(INT16_MAX) / (float)peak;
		gain = (gain > 1.0f) ? 1.0f : gain;

		if (gain != 1.0f) {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->record_pcm[j] = lrintf((float)pcm32[j] * gain);
		} else {
			for (j = 0; j < channels * ONE_OPUS_FRAME; j++)
				pctx->record_pcm[j] = pcm32[j];
		}

		/* write data to file */
		if (pctx->record == REC_WAVE) {
			wave_write_data(pctx->record_pcm, channels, ONE_OPUS_FRAME, &pctx->record_bytes_written, pctx->record_fp);
		} else {
			oggopus_write_data(pctx->record_pcm, ONE_OPUS_FRAME, pctx->record_enc, &pctx->record_oss,
			                   &pctx->record_packetno, 0, pctx->record_fp);
		}
	}

	return n+m;
}


int pc_engine_init(struct pc_context *pctx, const char *nick, int local_port, int bitrate,
                   const char *audiodevice_playback, const char *audiodevice_capture,
                   int autoaccept_calls, struct pollfd *pfds, unsigned int *nfds, int verbose)
{
	int ret, error, stdinfd;
	struct sockaddr_in local_addr;

	snprintf(msgbuf, MBS, "parole-conference-engine %s", PC_ENGINE_VERSION);
	logstr(msgbuf);

	/* clear parole-conference context (it'll clear nick, options and array of nodes) */
	memset(pctx, 0, sizeof(*pctx));
	strncpy(pctx->nick, nick, NICKLEN-1);
	pctx->local_port           = local_port;
	pctx->autoaccept_calls     = autoaccept_calls;
	pctx->verbose              = verbose;
	pctx->audiodevice_playback = audiodevice_playback;
	pctx->audiodevice_capture  = audiodevice_capture;
	pctx->micgain              = 1.0f;
	pctx->bitrate              = bitrate;

	/* if no nick was specified, choose a random one */
	if (pctx->nick[0] == '\0') {
		int i;
		memcpy(pctx->nick, "guest", 5);
		for (i = 0; i < 3; i++)
			pctx->nick[i+5] = '0' + (rand() % 10);

		snprintf(msgbuf, MBS, "No nick was specified, setting nick to '%s'", pctx->nick);
		msgbook_enqueue(&mb0, MB_TYPE_WARNING, MODULE, msgbuf);
	}

	/* get stdin fd */
	stdinfd = fileno(stdin);
	if (stdinfd == -1) {
		snprintf(msgbuf, MBS, "fileno(): %s", strerror(errno));
		goto engine_setup_fail;
	}
	/* create socket */
	pctx->sockfd = socket(PF_INET, SOCK_DGRAM, 0);
	if (pctx->sockfd == -1) {
		snprintf(msgbuf, MBS, "socket(): %s", strerror(errno));
		goto engine_setup_fail;
	}
	/* name socket (bind it to local_port) */
	local_addr.sin_family      = AF_INET;
	local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	local_addr.sin_port        = htons(pctx->local_port);
	ret = bind(pctx->sockfd, (struct sockaddr *)&local_addr, sizeof(local_addr));
	if (ret == -1) {
		snprintf(msgbuf, MBS, "bind(): %s", strerror(errno));
		goto engine_setup_fail;
	}

	/* file descriptor for stdin  is at pfds[fd_idx_stdin] */
	*nfds                     = 0;
	fd_idx_stdin              = *nfds;
	pfds[fd_idx_stdin].fd     = stdinfd;
	pfds[fd_idx_stdin].events = POLLIN;
	*nfds                    += 1;

	/* file descriptor for socket is at pfds[fd_idx_udpsocket] */
	fd_idx_udpsocket              = *nfds;
	pfds[fd_idx_udpsocket].fd     = pctx->sockfd;
	pfds[fd_idx_udpsocket].events = POLLIN;
	*nfds                        += 1;


	/* init alsa audio capture */
	fd_idx_audiocapture = *nfds;
	pctx->audiodevice_capture = pctx->audiodevice_capture ? pctx->audiodevice_capture : pctx->audiodevice_playback;
	ret = audio_init(&pctx->ad_capture, pctx->audiodevice_capture, SND_PCM_STREAM_CAPTURE,
		             samplerate, channels, pfds+*nfds, pctx->verbose);
	if (ret == -1)
		return -1;
	*nfds += pctx->ad_capture.nfds;

	/* init alsa audio playback */
	fd_idx_audioplayback = *nfds;
	ret = audio_init(&pctx->ad_playback, pctx->audiodevice_playback, SND_PCM_STREAM_PLAYBACK,
		             samplerate, channels, pfds+*nfds, pctx->verbose);
	if (ret == -1)
		return -1;
	/* we won't poll this device */
	/* *nfds += pctx->ad_playback.nfds; */


	/* create encoder state */
	logstr(opus_get_version_string());
	pctx->enc = opus_encoder_create(samplerate, channels, OPUS_APPLICATION_VOIP, &error);
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		goto engine_setup_fail;
	}
#if 0
	/* set maximum bandwidth */
	pctx->bandwidth = OPUS_BANDWIDTH_WIDEBAND;
	error = opus_encoder_ctl(pctx->enc, OPUS_SET_MAX_BANDWIDTH(pctx->bandwidth));
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		goto engine_setup_fail;
	}
#endif
	/* set bitrate */
	error = opus_encoder_ctl(pctx->enc, OPUS_SET_BITRATE(pctx->bitrate));
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		goto engine_setup_fail;
	}
	/* set signal type */
	error = opus_encoder_ctl(pctx->enc, OPUS_SET_SIGNAL(OPUS_SIGNAL_VOICE));
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		goto engine_setup_fail;
	}
	/* set packet loss */
	error = opus_encoder_ctl(pctx->enc, OPUS_SET_PACKET_LOSS_PERC(1));
	if (error != OPUS_OK) {
		snprintf(msgbuf, MBS, "opus error: %s", opus_strerror(error));
		goto engine_setup_fail;
	}

	/* generate tones */
	generate_tone(&pctx->tone);
	generate_calltone(&pctx->calltone);
	generate_ringtone(&pctx->ringtone);

	return 0;

engine_setup_fail:
	msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
	return -1;
}

void pc_engine_fix_revents(struct pc_context *pctx, struct pollfd *pfds)
{
	audio_poll_descriptors_revents(&pctx->ad_capture);
	pfds[fd_idx_audiocapture].revents = pctx->ad_capture.revents;
}

int pc_engine_network_ready(struct pc_context *pctx)
{
	struct udp_packet udp_packet;
	const uint8_t   *d;
	size_t           l;
	uint16_t packet_type;

	/* read udp_packet */
	receive_udp_packet(pctx->sockfd, &udp_packet);
	pctx->bytes_in_per_sec += udp_packet.len;

	if (udp_packet.len < 2)
		return RES_ERROR;

	d = udp_packet.data;
	l = udp_packet.len;
	packet_type = read_be16(d); d+=2; l-=2;

	/* if we have an encrypted packet we need to decrypt it */
	if (packet_type == UDP_PACKET_TYPE_ENCRYPTED) {
		/* if encryption is disable do not try to decrypt the packet */
		if (!pctx->encrypt) {
			if (pctx->verbose >= 2)
				msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, "Encrypted packet received, but encryption is disabled");
			return RES_ERROR;
		}
		if (l < 8) {
			msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "Invalid ENCRYPTED packet received");
			return RES_ERROR;
		} else {
			uint32_t iv[2];
			uint8_t  padding_size;
			struct udp_packet udp_packet_dec;

			/* read initialization vector */
			iv[0] = read_be32(d); d+=4; l-=4;
			iv[1] = read_be32(d); d+=4; l-=4;

			if (l > 0 && (l%8 == 0)) {
				/* copy payload to new packet */
				udp_packet_dec.addr = udp_packet.addr;
				memcpy(udp_packet_dec.data, d, l);
				udp_packet_dec.len = l;

				/* decrypt payload */
				xtea_decrypt_buffer_cbc(udp_packet_dec.data, udp_packet_dec.len, pctx->key, iv);
				padding_size = udp_packet_dec.data[udp_packet_dec.len - 1];
				udp_packet_dec.len -= padding_size;

				/* process packet */
				return process_packet(pctx, &udp_packet_dec);
			} else {
				return RES_ERROR;
			}
		}
	} else {
		if (pctx->encrypt) {
			if (pctx->verbose >= 2)
				msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, "Clear packet received, but encryption is enabled");
			return RES_ERROR;
		}
		/* process packet */
		return process_packet(pctx, &udp_packet);
	}
}

void pc_engine_step(struct pc_context *pctx)
{
	int i;
	struct slist *p;

	/* age callers */
	p = pctx->callers;
	while (p) {
		struct slist *pnext = p->next;
		struct pc_caller *cl = p->data;

		if (cl->timeout <= 0) {
			free(cl);
			pctx->callers = slist_remove_element(pctx->callers, p);
		} else {
			cl->timeout--;
		}
		p = pnext;
	}

	/* decrement timeout counter, delete node after 6s without receiving packets */
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i]) {
			if (pctx->nodes[i]->timeout <= 0)
				node_delete(pctx->nodes, i, REASON_TIMEOUT, &pctx->tone, &pctx->events);
			else
				pctx->nodes[i]->timeout--;
		}
	}

	/* send call packets periodically */
	if (pctx->call) {
		struct pc_sendrecipe *sr = pctx->call;
		int idx;

		idx = node_get_idx(pctx->nodes, sr->addr);
		if (idx == -1 || sr->packets_sent >= sr->resend_maxnum) {
			struct pc_event *ev;

			if (idx != -1)
				node_delete(pctx->nodes, idx, REASON_LEFT, &pctx->tone, &pctx->events);

			/* push event */
			ev = event_push(&pctx->events, event_type_calling_aborted);
			ev->addr = sr->addr;

			free(pctx->call);
			pctx->call = NULL;
		} else {
			/* reset timeout counter */
			pctx->nodes[idx]->timeout = TIMEOUT;

			if (sr->resend_watchdog <= 0) {
				send_udp_packet_enc(pctx, &sr->udp_packet);

				sr->resend_watchdog = sr->resend_interval;
				sr->packets_sent++;
			} else {
				sr->resend_watchdog--;
			}
		}
	}
#if 0
	/* packets to send periodically */
	p = pctx->sendrecipes;
	while (p) {
		struct slist *pnext = p->next;
		struct pc_sendrecipe *sr = p->data;
		int idx;

		idx = node_get_idx(pctx->nodes, sr->addr);
		if (idx == -1 || sr->packets_sent >= sr->resend_maxnum) {
			free(sr);
			pctx->sendrecipes = slist_remove_element(pctx->sendrecipes, p);
		} else {
			/* reset timeout counter */
			pctx->nodes[idx]->timeout = TIMEOUT;

			if (sr->resend_watchdog <= 0) {
				send_udp_packet_enc(pctx, &sr->udp_packet);

				sr->resend_watchdog = sr->resend_interval;
				sr->packets_sent++;
				if (pctx->verbose >= 2) {
					snprintf(msgbuf, MBS, "Resent packet for %s:%hu",
							 inet_ntoa(sr->udp_packet.addr.sin_addr), ntohs(sr->udp_packet.addr.sin_port));
					msgbook_enqueue(&mb0, MB_TYPE_DEBUG, MODULE, msgbuf);
				}
			} else {
				sr->resend_watchdog--;
			}
		}
		p = pnext;
	}
#endif

	/* connect packets or table packets might have gotten lost, so every
	   1s send out call requests to those who don't know about me,
	   eventually using relay to make sure packets get delivered */
	if ((1+pctx->ic_talk+pctx->ic_mute) % 50 == 0)
		connect_new_nodes(pctx, 0);

	/* update stats */
	if (pctx->ic_total % 50 == 0) {
		pctx->packets_received_g = 0;
		pctx->packets_lost_g     = 0;
		for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
			if (pctx->nodes[i]) {
				pctx->packets_received_g += pctx->nodes[i]->packets_received;
				pctx->packets_lost_g     += pctx->nodes[i]->packets_lost;
			}
		}
		pctx->bw_upload   = pctx->bytes_out_per_sec / 1024.0f;
		pctx->bw_download = pctx->bytes_in_per_sec  / 1024.0f;
		pctx->bytes_out_per_sec = 0;
		pctx->bytes_in_per_sec  = 0;
	}
}

void pc_engine_audio_ready(struct pc_context *pctx)
{
	snd_pcm_sframes_t ret_frames;

	/* read ad_capture.period_size frames of audio */
	ret_frames = audio_read(&pctx->ad_capture, 1);
	if (ret_frames < 0) {
		int ret;
		/* overrun, let's handle it and try again */
		pctx->overruns++;
		ret = audio_recover(&pctx->ad_capture, ret_frames, 1);
		if (ret == 0) {
			ret = audio_start(&pctx->ad_capture);
			if (ret == -1)
				die("audio_start() failed", 1);
			ret_frames = audio_read(&pctx->ad_capture, 0);
			if (ret_frames < 0)
				die("audio_read() failed", 1);
		} else {
			die("audio_recover() on capture failed", 1);
		}
	}

	/* make sure we have read an entire frame */
	if ((snd_pcm_uframes_t)pctx->ad_capture.frames != pctx->ad_capture.alsabuffersize) {
		snprintf(msgbuf, MBS, "audio_read() only read %ld frames instead of %lu",
			     pctx->ad_capture.frames, pctx->ad_capture.alsabuffersize);
		die(msgbuf, 1);
	}
	pctx->lb_pcmlen = (unsigned int)pctx->ad_capture.frames;

	if (!pctx->micmute) {
		int32_t ipeak;
		float peak_percent;
		double rms2;

		/* amplify if necessary */
		if (pctx->micgain != 1.0f) {
			unsigned int i;
			int32_t vl, vr;
			for (i = 0; i < pctx->ad_capture.channels * (unsigned int)pctx->ad_capture.frames; i += 2) {
				vl = lrintf((float)pctx->ad_capture.alsabuffer[i  ] * pctx->micgain);
				vr = lrintf((float)pctx->ad_capture.alsabuffer[i+1] * pctx->micgain);
				vl = (vl < INT16_MIN) ? INT16_MIN : ((vl > INT16_MAX) ? INT16_MAX : vl);
				vr = (vr < INT16_MIN) ? INT16_MIN : ((vr > INT16_MAX) ? INT16_MAX : vr);
				pctx->lb_pcm[i  ] = vl;
				pctx->lb_pcm[i+1] = vr;
			}
		} else {
			/* copy captured audio to loopback buffer */
			memcpy(pctx->lb_pcm, pctx->ad_capture.alsabuffer,
				   pctx->ad_capture.channels * pctx->ad_capture.frames * sizeof(*pctx->ad_capture.alsabuffer));
		}

		/* find peak in captured audio */
		ipeak = find_peak_s16_2ch(pctx->lb_pcm, channels * pctx->lb_pcmlen);
		peak_percent = 100.0f * (float)ipeak / (float)(INT16_MAX+1);
		pctx->peak_percent = (pctx->peak_percent < peak_percent) ? peak_percent : pctx->peak_percent;

		/* dbSPL */
		pctx->sos   += sum_of_squares(pctx->lb_pcm, channels * pctx->lb_pcmlen);
		pctx->sos_N += channels * pctx->lb_pcmlen;
		rms2 = (double)(pctx->sos) / (double)pctx->sos_N;
		pctx->dBSPL = 10.0 * log10(1.0 /* to avoid -inf */ + rms2) - 90.3 /* 20*log10(1/32768.0) */;
	} else {
		/* insert silence in loopback buffer */
		memset(pctx->lb_pcm, 0, sizeof(pctx->lb_pcm));
	}

	/* if there are no nodes connected, there is nothing to do */
	if (node_get_count(pctx->nodes) == 0) {
		pctx->ic_talk = 0;
		pctx->ic_mute = 0;
	} else {
		struct udp_packet udp_packet;

		if (pctx->micmute) {
			if ((1+pctx->ic_total) % 50 == 0) { /* nop packets only every 1s */
				/* forge udp nop packet */
				packet_nop(&udp_packet);

				/* send udp packet */
				send_udp_packet_enc_to_all(pctx, &udp_packet);
			}

			/* increment mute frame counter */
			pctx->ic_mute++;
		} else {
			uint8_t    opus_enc_packet[4000];
			opus_int32 opus_enc_packetlen;

			/* encoding */
			opus_enc_packetlen = opus_encode(pctx->enc, pctx->lb_pcm, pctx->lb_pcmlen, opus_enc_packet, 4000);
			if (opus_enc_packetlen < 0) {
				snprintf(msgbuf, MBS, "opus_encode() failed: %s", opus_strerror(opus_enc_packetlen));
				msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
			} else {
				/* forge udp audio packet */
				packet_audio(&udp_packet, pctx->ic_talk, opus_enc_packet, opus_enc_packetlen);

				/* send udp packet */
				send_udp_packet_enc_to_all(pctx, &udp_packet);
			}

			/* increment talk frame counter */
			pctx->ic_talk++;
		}
	}

	/* mix audio */
	if (mix_audio(pctx)) {
		/* playback decoded data */
		ret_frames = audio_write(&pctx->ad_playback, pctx->pcm, pctx->pcmlen, 1);
		if (ret_frames < 0) {
			int ret;
			/* underrun, let's handle it and try again */
			pctx->underruns++;
			ret = audio_recover(&pctx->ad_playback, ret_frames, 1);
			if (ret == 0) {
				ret_frames = audio_write(&pctx->ad_playback, pctx->pcm, pctx->pcmlen, 0);
				if (ret_frames < 0)
					die("audio_write() failed", 1);
			} else {
				die("audio_recover() on playback failed", 1);
			}
		}
	}

	/* increment total frame counter */
	pctx->ic_total++;
}

void pc_engine_goodbye(struct pc_context *pctx)
{
	struct udp_packet udp_packet;

	packet_bye(&udp_packet);
	send_udp_packet_enc_to_all(pctx, &udp_packet);
}

void pc_engine_cleanup(struct pc_context *pctx)
{
	int i;

	/* stop recording if necessary */
	pc_engine_cmd_stop_recording(pctx);

	close(pctx->sockfd);
	audio_stop(&pctx->ad_capture);
	audio_close(&pctx->ad_capture);
	audio_stop(&pctx->ad_playback);
	audio_close(&pctx->ad_playback);
	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		if (pctx->nodes[i])
			node_delete(pctx->nodes, i, REASON_LEFT, &pctx->tone, &pctx->events);
	}

	free(pctx->ringtone.pcm);
	if (pctx->call)
		free(pctx->call);

	/* destroy encoder state */
	opus_encoder_destroy(pctx->enc);
}


void pc_engine_cmd_print_help(void)
{
	msgbook_enqueue(&mb0, MB_TYPE_INFO, MODULE,
"help:\n"
"command          description            shortcut  range\n"
"  /h             print help\n"
"  /i             print info             [F4]\n"
"  /m             toggle mic mute        [F5]\n"
"  /l             toggle audio loopback  [F6]\n"
"  /[r,R]         toggle call recording\n"
"  /g gain        set software mic gain            ]0,10]\n"
"  /G n gain      set node n gain in dB            [-40,40]\n"
"  /b bitrate     set bitrate                      [6000,512000]\n"
"  /k n           kill node n\n"
"  /c host[:port] call host\n"
"  /e             toggle encryption\n"
"  /x key         set encryption key\n"
"  /q             quit                   [CTRL-C]"
	);
}

void pc_engine_cmd_print_info(struct pc_context *pctx)
{
	unsigned int i;
	snd_pcm_sframes_t ret_frames;

	update_bitrate_and_bandwidth(pctx);

	/* read available frames */
	ret_frames = audio_avail(&pctx->ad_capture, 1);
	if (ret_frames < 0)
		pctx->ad_capture.avail = pctx->ad_capture.buffer_size;
	ret_frames = audio_avail(&pctx->ad_playback, 1);
	if (ret_frames < 0) {
		pctx->ad_playback.avail = 0;
		pctx->underruns++;
		audio_recover(&pctx->ad_playback, ret_frames, 1);
	}

	snprintf(msgbuf, MBS,
"info:\n"
"  capture:   rb[%3.0f%%]   overrun[%2d]  gain[%.2f]  mute[%s]\n"
"  playback:  rb[%3.0f%%]  underrun[%2d]  loopback[%s]\n"
"  stream:    bitrate[%5d bit/s]  bandwidth[%d Hz]\n"
"  xtea:      enc[%s]  key[%08x%08x%08x%08x]\n"
"  nodes: %u\n"
"    -:%15s:%-5hu %12s",
	100.0f * pctx->ad_capture.avail/(float)pctx->ad_capture.buffer_size, pctx->overruns, pctx->micgain, pctx->micmute ? " on" : "off",
	100.0f * pctx->ad_playback.avail/(float)pctx->ad_playback.buffer_size, pctx->underruns, pctx->lb ? " on" : "off",
	pctx->bitrate, pctx->bandwidth_hz,
	pctx->encrypt ? " on" : "off", pctx->key[0], pctx->key[1], pctx->key[2], pctx->key[3],
	node_get_count(pctx->nodes),
	pctx->external_ip[0] ? pctx->external_ip : "localhost", pctx->local_port, pctx->nick);

	for (i = 0; i < MAX_NUMBER_OF_NODES; i++) {
		size_t l;
		if (pctx->nodes[i]) {
			l = strlen(msgbuf);
			snprintf(msgbuf+l, MBS-l, "\n    %u:%15s:%-5hu %12s  pl[%3d/%4d]  pcm[%3ums]  %+.0fdB%s",
			         i, inet_ntoa(pctx->nodes[i]->addr.sin_addr),
			         ntohs(pctx->nodes[i]->addr.sin_port), pctx->nodes[i]->nick,
			         pctx->nodes[i]->packets_lost,
			         pctx->nodes[i]->packets_lost + pctx->nodes[i]->packets_received,
			         20 * pctx->nodes[i]->nb_frames,
			         (float)pctx->nodes[i]->gain_Q8_dB / 256.0f,
			         pctx->nodes[i]->nb_frames ? "" : "  MUTE");
		}
	}
	msgbook_enqueue(&mb0, MB_TYPE_INFO, MODULE, msgbuf);
}

void pc_engine_cmd_toggle_mute(struct pc_context *pctx)
{
	pctx->micmute = (pctx->micmute == 0) ? 1 : 0;
	event_push(&pctx->events, event_type_mute);
}

void pc_engine_cmd_toggle_loopback(struct pc_context *pctx)
{
	pctx->lb = (pctx->lb == 0) ? 1 : 0;
	event_push(&pctx->events, event_type_loopback);
}

void pc_engine_cmd_stop_recording(struct pc_context *pctx)
{
	if (pctx->record == REC_WAVE) {
		wave_stop_recording(pctx->record_bytes_written, pctx->record_fp);
		pctx->record = 0;
		event_push(&pctx->events, event_type_recording);
	} else if (pctx->record == REC_OPUS) {
		memset(pctx->record_pcm, 0, sizeof(pctx->record_pcm));
		oggopus_write_data(pctx->record_pcm, ONE_OPUS_FRAME, pctx->record_enc, &pctx->record_oss,
		                   &pctx->record_packetno, 1, pctx->record_fp);
		oggopus_stop_recording(&pctx->record_enc, &pctx->record_oss, pctx->record_fp);
		pctx->record = 0;
		event_push(&pctx->events, event_type_recording);
	}
}

int pc_engine_cmd_toggle_recording(struct pc_context *pctx, int rec_format)
{
	if (rec_format != REC_WAVE && rec_format != REC_OPUS)
		rec_format = REC_OPUS;

	if (pctx->record == 0) {
		time_t now;
		struct tm *t;

		now = time(NULL);
		t = localtime(&now);

		if (t)
			snprintf(pctx->record_filename, 64, "rec_%04d%02d%02d_%02d%02d%02d",
			         t->tm_year+1900, t->tm_mon+1, t->tm_mday,
			         t->tm_hour, t->tm_min, t->tm_sec);
		else
			snprintf(pctx->record_filename, 64, "rec_%u", (unsigned int)now);

		if (rec_format == REC_WAVE) {
			strcat(pctx->record_filename, ".wav");
			pctx->record_fp = wave_start_recording(pctx->record_filename, samplerate, channels);
		} else {
			strcat(pctx->record_filename, ".opus");
			pctx->record_fp = oggopus_start_recording(pctx->record_filename, samplerate, channels,
			                                          &pctx->record_enc, &pctx->record_oss);
		}

		if (pctx->record_fp) {
			pctx->record               = rec_format;
			pctx->record_bytes_written = 0;
			pctx->record_packetno      = 2;
			event_push(&pctx->events, event_type_recording);
			return 0;
		} else {
			return -1;
		}
	} else {
		pc_engine_cmd_stop_recording(pctx);
		return 0;
	}
}

int pc_engine_cmd_set_micgain(struct pc_context *pctx, float gain)
{
	if (0.0f < gain && gain <= 10.0f) {
		pctx->micgain = gain;
		event_push(&pctx->events, event_type_micgain);
		return 0;
	} else {
		return -1;
	}
}

int pc_engine_cmd_set_bitrate(struct pc_context *pctx, int bitrate)
{
	if (6000 <= bitrate && bitrate <= 512000) {
		opus_encoder_ctl(pctx->enc, OPUS_SET_BITRATE(bitrate));
		update_bitrate_and_bandwidth(pctx);
		event_push(&pctx->events, event_type_bitrate);
		return 0;
	} else {
		return -1;
	}
}

void pc_engine_cmd_toggle_encryption(struct pc_context *pctx)
{
	pctx->encrypt = (pctx->encrypt == 0) ? 1 : 0;
	event_push(&pctx->events, event_type_encryption);
}

static uint8_t hex2num(uint8_t c)
{
	if ('0' <= c && c <= '9')
		return (c - '0');
	else if ('a' <= c && c <= 'f')
		return (c - 'a' + 10);
	else if ('A' <= c && c <= 'F')
		return (c - 'A' + 10);
	else
		return 0;
}

void pc_engine_cmd_set_key(struct pc_context *pctx, const char hexkey[32])
{
	int i;

	/* convert key from hex and store it to key[] array */
	memset(pctx->key, 0, sizeof(pctx->key));
	for (i = 0; i < 32; i++)
		pctx->key[i/8] |= hex2num(hexkey[i]) << 4*(7 - i%8);

	event_push(&pctx->events, event_type_setkey);
}

int pc_engine_cmd_set_nodegain(struct pc_context *pctx, int idx, float gain_dB)
{
	if (0 <= idx && idx < MAX_NUMBER_OF_NODES && pctx->nodes[idx] &&
	    -40.0f <= gain_dB && gain_dB <= 40.0f) {
		struct pc_event *ev;

		pctx->nodes[idx]->gain_Q8_dB = lrintf(gain_dB * 256.0f);
		opus_decoder_ctl(pctx->nodes[idx]->dec, OPUS_SET_GAIN(pctx->nodes[idx]->gain_Q8_dB));
		ev = event_push(&pctx->events, event_type_nodegain);
		ev->addr = pctx->nodes[idx]->addr;
		memcpy(ev->nick, pctx->nodes[idx]->nick, NICKLEN);
		ev->f = gain_dB;
		return 0;
	} else {
		return -1;
	}
}

int pc_engine_cmd_kill_node(struct pc_context *pctx, int idx)
{
	if (0 <= idx && idx < MAX_NUMBER_OF_NODES && pctx->nodes[idx]) {
		node_delete(pctx->nodes, idx, REASON_LEFT, &pctx->tone, &pctx->events);
		return 0;
	}
	return -1;
}

int pc_engine_cmd_call_node(struct pc_context *pctx, const char *host, int port)
{
	struct udp_packet udp_packet;
	struct sockaddr_in addr;
	struct pc_event *ev;
	struct pc_sendrecipe *sr;
	int ret, idx;

	if (pctx->call) {
		msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "We're already calling a host");
		return -1;
	}

	ret = resolve_address(&addr, host, port, pctx->verbose);
	if (ret == -1)
		return -1;

	idx = node_get_idx(pctx->nodes, addr);
	if (idx == -1) {
		idx = node_add(pctx->nodes, addr, NULL, &pctx->tone, &pctx->events);
		if (idx == -1)
			return -1;
		pctx->nodes[idx]->status = STATUS_CALLING;
	}

	udp_packet.addr = addr;
	packet_call(&udp_packet, pctx->nick);

	/* insert packet in call sendrecipe */
	sr = xcalloc(1, sizeof(*sr));
	sr->addr            = addr;
	sr->udp_packet      = udp_packet;
	sr->resend_interval = 50; /* every second */
	sr->resend_maxnum   = 60; /* 60 times */
	sr->resend_watchdog = 0;  /* send as soon as possible */
	sr->packets_sent    = 0;
	pctx->call          = sr;

	/* push event */
	ev = event_push(&pctx->events, event_type_calling_in_progress);
	ev->addr = addr;

	return 0;
}

void pc_engine_cmd_send_chat(struct pc_context *pctx, const char *buf, size_t len)
{
	struct udp_packet udp_packet;
	struct pc_event *ev;

	/* push event */
	ev = event_push(&pctx->events, event_type_chat);
	/* ev->addr = ; */
	memcpy(ev->nick, pctx->nick, NICKLEN);
	ev->data = xmalloc(len);
	memcpy(ev->data, buf, len);

	packet_chat(&udp_packet, buf, len);
	send_udp_packet_enc_to_all(pctx, &udp_packet);
}

void pc_engine_cmd_accept_call(struct pc_context *pctx)
{
	if (pctx->callers) {
		struct pc_caller *cl;
		int idx;
		struct pc_event *ev;
		struct udp_packet udp_packet_response;

		cl = pctx->callers->data;

		/* push event */
		ev = event_push(&pctx->events, event_type_chose_accepted);
		ev->addr = cl->addr;
		memcpy(ev->nick, cl->nick, NICKLEN);

		idx = node_add(pctx->nodes, cl->addr, cl->nick, &pctx->tone, &pctx->events);
		if (idx == -1) {
			/* send refuse */
			udp_packet_response.addr = cl->addr;
			packet_refuse(&udp_packet_response);
			send_udp_packet_enc(pctx, &udp_packet_response);
			return;
		}
		pctx->nodes[idx]->status = STATUS_OK;

		/* send table */
		udp_packet_response.addr = cl->addr;
		packet_table(&udp_packet_response, pctx->nick, pctx->nodes, idx);
		send_udp_packet_enc(pctx, &udp_packet_response);

		free(cl);
		pctx->callers = slist_remove_head(pctx->callers);
	}
}

void pc_engine_cmd_refuse_call(struct pc_context *pctx)
{
	if (pctx->callers) {
		struct pc_caller *cl;
		struct pc_event *ev;
		struct udp_packet udp_packet_response;

		cl = pctx->callers->data;

		/* push event */
		ev = event_push(&pctx->events, event_type_chose_refused);
		ev->addr = cl->addr;
		memcpy(ev->nick, cl->nick, NICKLEN);

		/* send refuse */
		udp_packet_response.addr = cl->addr;
		packet_refuse(&udp_packet_response);
		send_udp_packet_enc(pctx, &udp_packet_response);

		free(cl);
		pctx->callers = slist_remove_head(pctx->callers);
	}
}

#undef MODULE
