/*
 * Copyright (C) 2013 Giorgio Vazzana
 *
 * This file is part of Seren.
 *
 * Seren is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Seren is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <poll.h>
#include <time.h>
#include <locale.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "pc-engine.h"
#include "common.h"
#include "msgbook.h"
#include "nc.h"
#include "audio.h"
#include "stun.h"
#include "config.h"

#define SERENDIR  ".seren"
#define LOGFILE   "seren.log"
#define CONFFILE  "seren.conf"
#define MODULE    "main"

static volatile sig_atomic_t keep_going;
static volatile sig_atomic_t window_resized;

static void signal_handler(int sig)
{
	if (sig == SIGINT)
		keep_going = 0;
	if (sig == SIGWINCH)
		window_resized = 1;
}

void my_die_cb(const char *msg, int exit_code)
{
	/* shut down ncurses if necessary */
	if (nctx.initialized) {
		endwin();
		nctx.initialized = 0;
	}

	mb0.print_cb = msgbook_print_cb_stderr;
	msgbook_flush(&mb0);

	logstr(msg);

	exit_code &= 0xFF;
}

void handle_events(struct pc_context *pctx)
{
	struct slist *p;

	p = pctx->events;
	while (p) {
		struct slist *pnext = p->next;
		struct pc_event *ev = p->data;

		switch (ev->type) {
		case event_type_mute:
			snprintf(msgbuf, MBS, "mute: %s", pctx->micmute ? "on" : "off");
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_loopback:
			snprintf(msgbuf, MBS, "loopback: %s", pctx->lb ? "on" : "off");
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_recording:
			snprintf(msgbuf, MBS, "recording: %s%s%s", pctx->record ? "on" : "off",
			         pctx->record ? ", file: " : "",
			         pctx->record ? pctx->record_filename : "");
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_micgain:
			snprintf(msgbuf, MBS, "mic gain: %.2f", pctx->micgain);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_bitrate:
			snprintf(msgbuf, MBS, "bitrate: %d bit/s", pctx->bitrate);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_encryption:
			snprintf(msgbuf, MBS, "encryption: %s", pctx->encrypt ? "on" : "off");
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_setkey:
			snprintf(msgbuf, MBS, "xtea key: %08x%08x%08x%08x", pctx->key[0], pctx->key[1], pctx->key[2], pctx->key[3]);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_nodegain:
			snprintf(msgbuf, MBS, "gain for node %s:%hu (%s) set to %.2f dB",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick, ev->f);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_node_delete:
			snprintf(msgbuf, MBS, "Node %s:%hu (%s) left%s",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick, ev->i ? " (timeout)" : " (call ended)");
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_node_add:
			snprintf(msgbuf, MBS, "Node %s:%hu (%s) joined the network",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_chat:
			msgbook_enqueue(&mb0, MB_TYPE_CHAT, ev->nick, (char *)ev->data);
			break;
		case event_type_calling_in_progress:
			snprintf(msgbuf, MBS, "Calling %s:%hu...",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_calling_aborted:
			snprintf(msgbuf, MBS, "Calling %s:%hu aborted",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port));
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_incoming_call:
			snprintf(msgbuf, MBS, "Node %s:%hu (%s) is calling, press F7 to accept, F8 to refuse",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_chose_refused:
			snprintf(msgbuf, MBS, "Call refused for %s:%hu (%s)",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_chose_accepted:
			snprintf(msgbuf, MBS, "Call accepted for %s:%hu (%s)",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_call_refused:
			snprintf(msgbuf, MBS, "Node %s:%hu (%s) refused the call",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		case event_type_call_accepted:
			snprintf(msgbuf, MBS, "Node %s:%hu (%s) accepted the call",
					 inet_ntoa(ev->addr.sin_addr), ntohs(ev->addr.sin_port),
					 ev->nick);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
			break;
		}

		free(ev->data);
		free(ev);
		pctx->events = slist_remove_element(pctx->events, p);

		p = pnext;
	}
}

static void usage()
{
	fputs(
"Usage: seren [option...]\n"
"\n"
"Options:\n"
"  -h             print help\n"
"  -v             increment verbosity level\n"
"  -t theme       set theme\n"
"                 0-simple, 1-dark, 2-clear     (1)\n"
"  -n nick        set nickname                  (login name)\n"
"  -p port        set udp listening port        (8110)\n"
"  -b bitrate     set bitrate in bit/s          (16000)\n"
"  -d audiodev    set playback audio device     (hw:0,0)\n"
"  -D audiodev    set capture audio device      (playback device)\n"
"  -a             autoaccept calls\n"
"  -S             disable STUN\n",
	stderr);
}

struct options {
	int   verbose;
	int   theme;
	char  nick[NICKLEN];
	int   local_port;
	int   bitrate;
	char *audiodevice_playback;
	char *audiodevice_capture;
	int   autoaccept_calls;
	int   disable_stun;
};

static void parse_conf_file(const char *filename, struct options *options)
{
	char buf[128];
	FILE *fp;

	fp = fopen(filename, "r");
	if (!fp) {
		/* create default conf file */
		FILE *fp_w = fopen(filename, "w");
		if (fp_w) {
			fputs(
"# Seren configuration file\n"
"# To activate an option uncomment the associated line\n"
"\n"
"#verbose          0\n"
"#theme            0\n"
"#nick             mynick\n"
"#port             8110\n"
"#bitrate          16000\n"
"#playback_dev     hw:0,0\n"
"#capture_dev      hw:0,0\n"
"#autoaccept_calls 0\n"
"#disable_stun     1\n",
			fp_w);
			fclose(fp_w);
		}
		return;
	}

	while (1) {
		char *pc, *name, *value;
		size_t len;

		/* read an entire line */
		pc = fgets(buf, 128, fp);
		if (pc == NULL)
			break;

		/* skip if it's a comment or does not end with '\n' */
		len = strlen(buf);
		if (buf[0] == '#' || buf[len-1] != '\n')
			continue;

		/* delete trailing '\n' */
		buf[len-1] = '\0';
		len--;

		/* skip leading spaces */
		while (*pc == ' ')
			pc++;
		name = pc;

		/* find space(s) that separates property name from value */
		pc = strchr(name, ' ');
		if (pc == NULL)
			continue;
		*pc = '\0';
		/* skip leading spaces */
		pc++;
		while (*pc == ' ')
			pc++;
		value = pc;
		/* delete trailing spaces */
		pc = strchr(value, ' ');
		if (pc)
			*pc = '\0';

		if (strcmp(name, "verbose") == 0) {
			options->verbose = atoi(value);
		} else if (strcmp(name, "theme") == 0) {
			options->theme = atoi(value);
		} else if (strcmp(name, "nick") == 0) {
			strncpy(options->nick, value, NICKLEN-1);
			options->nick[NICKLEN-1] = '\0';
		} else if (strcmp(name, "port") == 0) {
			options->local_port = atoi(value);
		} else if (strcmp(name, "bitrate") == 0) {
			options->bitrate = atoi(value);
		} else if (strcmp(name, "playback_dev") == 0) {
			free(options->audiodevice_playback);
			options->audiodevice_playback = xmalloc(strlen(value)+1);
			strcpy(options->audiodevice_playback, value);
		} else if (strcmp(name, "capture_dev") == 0) {
			free(options->audiodevice_capture);
			options->audiodevice_capture = xmalloc(strlen(value)+1);
			strcpy(options->audiodevice_capture, value);
		} else if (strcmp(name, "autoaccept_calls") == 0) {
			options->autoaccept_calls = (value[0] == '1') ? 1 : 0;
		} else if (strcmp(name, "disable_stun") == 0) {
			options->disable_stun = (value[0] == '1') ? 1 : 0;
		}
	}
	fclose(fp);
}


int main(int argc, char *argv[])
{
	int                  ret, opt, pollret, keep_going_capture;
	struct pollfd        pfds[6];
	unsigned int         nfds;
	struct options       options;
	char                *home, path[128];
	struct sigaction     act;
	struct pc_context    pctx;
	struct udp_packet    udp_packet;


	/* set locale and print banner */
	die_cb = NULL;
	srand(time(NULL));
	setlocale(LC_ALL, "");
	fprintf(stderr, "%s, ver. %s\n\n", PACKAGE_NAME, PACKAGE_VERSION);

	/* make log/conf directory */
	home = getenv("HOME");
	if (home == NULL || home[0] == '\0' || strlen(home) > 100)
		die("$HOME is not defined, empty or too long", 1);
	strcpy(path, home);
	strcat(path, "/"SERENDIR);
	mkdir(path, S_IRUSR|S_IWUSR|S_IXUSR | S_IRGRP|S_IWGRP|S_IXGRP);

	/* set default options */
	memset(&options, 0, sizeof(options));
	strncpy(options.nick, getlogin(), NICKLEN-1);
	options.theme                = 1;
	options.local_port           = 8110;
	options.bitrate              = 16000;
	options.audiodevice_playback = strdup("hw:0,0");
	options.audiodevice_capture  = NULL;

	/* read options from conf file */
	home = getenv("HOME");
	strcpy(path, home);
	strcat(path, "/"SERENDIR"/"CONFFILE);
	parse_conf_file(path, &options);

	/* parse command line options */
	while ((opt = getopt(argc, argv, "hvt:n:p:b:d:D:aS")) != -1) {
		switch (opt) {
		case 'h':
			usage();
			exit(0);
		case 'v':
			options.verbose++;
			break;
		case 't':
			options.theme = atoi(optarg);
			break;
		case 'n':
			strncpy(options.nick, optarg, NICKLEN-1);
			options.nick[NICKLEN-1] = '\0';
			break;
		case 'p':
			options.local_port = atoi(optarg);
			break;
		case 'b':
			options.bitrate = atoi(optarg);
			break;
		case 'd':
			free(options.audiodevice_playback);
			options.audiodevice_playback = xmalloc(strlen(optarg)+1);
			strcpy(options.audiodevice_playback, optarg);
			break;
		case 'D':
			free(options.audiodevice_capture);
			options.audiodevice_capture = xmalloc(strlen(optarg)+1);
			strcpy(options.audiodevice_capture, optarg);
			break;
		case 'a':
			options.autoaccept_calls = 1;
			break;
		case 'S':
			options.disable_stun = 1;
			break;
		default:
			fprintf(stderr, "\n");
			usage();
			exit(1);
		}
	}

#if 0
	fprintf(stderr, "verbose=%d, theme=%d, nick='%s', local_port=%d, bitrate=%d, audiodevice_playback='%s', "
	                "audiodevice_capture='%s', autoaccept_calls=%d, disable_stun=%d\n",
	        options.verbose, options.theme, options.nick, options.local_port, options.bitrate,
	        options.audiodevice_playback, options.audiodevice_capture, options.autoaccept_calls, options.disable_stun);
#endif

	/* init msgbook */
	msgbook_init(&mb0);
	mb0.print_cb = msgbook_print_cb_stderr;

	/* open log file */
	home = getenv("HOME");
	strcpy(path, home);
	strcat(path, "/"SERENDIR"/"LOGFILE);
	fplog = fopen(path, "a");
	if (!fplog) {
		snprintf(msgbuf, MBS, "Cannot open log file '%s'", path);
		die(msgbuf, 1);
	}
	logstr(PACKAGE_STRING);

	/* init parole-conference engine */
	ret = pc_engine_init(&pctx, options.nick, options.local_port, options.bitrate,
	                     options.audiodevice_playback, options.audiodevice_capture,
	                     options.autoaccept_calls, pfds, &nfds, options.verbose);
	if (ret == -1)
		die("pc_engine_init() failed", 1);

	/* init ncurses */
	memset(&nctx, 0, sizeof(nctx));
	ret = nc_init(PACKAGE_VERSION, pctx.nick, options.theme);
	if (ret == -1)
		die("nc_init() failed", 1);
	/* set callbacks */
	mb0.print_cb = nc_print_cb;
	die_cb = my_die_cb;

	/* install signal handler for window resize */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	ret = sigaction(SIGWINCH, &act, NULL);
	if (ret == -1)
		die("sigaction() failed", 1);
	window_resized = 0;

	/* welcome messages */
	msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, "Welcome! Enter /h for help");
	snprintf(msgbuf, MBS, "Listening for incoming calls on udp port %d...", pctx.local_port);
	msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);

	/* install signal handler */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	ret = sigaction(SIGINT, &act, NULL);
	if (ret == -1)
		die("sigaction() failed", 1);
	keep_going = 1;

	/* get external ip address */
	if (!options.disable_stun) {
		char *ip = stun_get_external_ip(options.verbose);
		if (ip)
			strcpy(pctx.external_ip, ip);
	}

	/* start audio capture */
	ret = audio_start(&pctx.ad_capture);
	if (ret == -1)
		die("audio_start() failed", 1);
	keep_going_capture = 1;


	/* main loop */
	while (keep_going || keep_going_capture) {
		/* poll: is audio_from_capture or udp_socket ready? */
		pollret = poll(pfds, nfds, -1);

		/* poll can fail and return -1 when a signal arrives. errno will be EINTR */
		if (pollret == -1) {
			if (errno == EINTR) {
				continue;
			} else {
				snprintf(msgbuf, MBS, "poll(): %s", strerror(errno));
				die(msgbuf, 1);
			}
		}
#if 0
		/* timeout should never happen really... */
		if (pollret == 0) {
			snprintf(msgbuf, MBS, "poll timeout");
			logstr(msgbuf);
			continue;
		}
#endif
		/* window resize */
		if (window_resized) {
			endwin();
			refresh();

			getmaxyx(stdscr, nctx.rows, nctx.cols);
			clear();
			refresh();

			wresize(nctx.win, nctx.rows - 4, nctx.cols);
			wclear(nctx.win);
			snprintf(msgbuf, MBS, "Window resized to %dx%d", nctx.cols, nctx.rows);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);

			clear_input_line(1);
			print_bars(PACKAGE_VERSION);
			print_prompt(pctx.nick);

			window_resized = 0;
		}


		/** UDP SOCKET **/
		if (pfds[fd_idx_udpsocket].revents & POLLERR)
			die("udp socket POLLERR", 1);

		/* process data from socket if ready */
		if (pfds[fd_idx_udpsocket].revents & POLLIN) {
			/* read and process packet */
			ret = pc_engine_network_ready(&pctx);
			handle_events(&pctx);
			continue;
		}


		/** STDIN **/
		if (pfds[fd_idx_stdin].revents & POLLIN) {
			int ch;
			size_t len;

			/* read an entire line, skipping special keys */
			ch = getch();
#if 0
			snprintf(msgbuf, MBS, "key: %u|0%o", ch, ch);
			msgbook_enqueue(&mb0, MB_TYPE_GUI, MODULE, msgbuf);
#endif
			switch (ch) {
			case 27: /* ESC */
			case KEY_RESIZE:
			case KEY_DC:
			case KEY_IC:
			case KEY_DOWN:
			case KEY_UP:
			case KEY_LEFT:
			case KEY_RIGHT:
			case KEY_NPAGE:
			case KEY_PPAGE:
			case KEY_HOME:
			case KEY_END:
			case KEY_ENTER:
				continue;
			case KEY_F(4):
				pc_engine_cmd_print_info(&pctx);
				continue;
			case KEY_F(5):
				pc_engine_cmd_toggle_mute(&pctx);
				continue;
			case KEY_F(6):
				pc_engine_cmd_toggle_loopback(&pctx);
				continue;
			case KEY_F(7):
				pc_engine_cmd_accept_call(&pctx);
				continue;
			case KEY_F(8):
				pc_engine_cmd_refuse_call(&pctx);
				continue;
			default:
				break;
			}

			if (ch == KEY_BACKSPACE) {
				if (nctx.wlinepos) {
					nctx.wlinepos--;

					/* if we have a continuation byte, erase 2 chars */
					if (nctx.wlinebuf[nctx.wlinepos][0] < 128)
						nctx.linepos--;
					else
						nctx.linepos -= 2;

					/* scroll text if necessary */
					if (nctx.wlinepos >= (nctx.cols - (strlen(pctx.nick)+2)) - 1) {
						unsigned int i;
						clear_input_line(0);
						print_prompt(pctx.nick);
						for (i = nctx.wlinepos - (nctx.cols - (strlen(pctx.nick)+2) - 1); i < nctx.wlinepos; i++) {
							addch(nctx.wlinebuf[i][0]);
							if (nctx.wlinebuf[i][0] >= 128)
								addch(nctx.wlinebuf[i][1]);
						}
						refresh();
					} else {
						int y, x;

						getyx(stdscr, y, x);
						move(y, x-1);
						addch(' ');
						move(y, x-1);
						refresh();
					}
				}
				continue;
			}

			/* limit max characters read */
			if (ch != '\n' && nctx.wlinepos >= 1000) {
				pctx.tone.tone_play = (pctx.tone.tone_play > 10*ONE_OPUS_FRAME) ? pctx.tone.tone_play : 10*ONE_OPUS_FRAME;
				continue;
			}

			/* scroll text */
			if (nctx.wlinepos >= (nctx.cols - (strlen(pctx.nick)+2)) - /* leave a space on the left */ 1) {
				unsigned int i;
				clear_input_line(0);
				print_prompt(pctx.nick);
				for (i = nctx.wlinepos - (nctx.cols - (strlen(pctx.nick)+2) - 1) + 1; i < nctx.wlinepos; i++) {
					addch(nctx.wlinebuf[i][0]);
					if (nctx.wlinebuf[i][0] >= 128)
						addch(nctx.wlinebuf[i][1]);
				}
				refresh();
			}

			/* insert char */
			if (ch < 128) {
				nctx.linebuf[nctx.linepos++] = ch;
				nctx.wlinebuf[nctx.wlinepos++][0] = ch;
				addch(ch);
				refresh();
				if (ch != '\n')
					continue;
			} else if ((ch & 0xe0) == 0xc0) { /* UTF-8 char encoded in 2 bytes */
				nctx.linebuf[nctx.linepos++] = ch;
				nctx.wlinebuf[nctx.wlinepos][0] = ch;
				addch(ch);
				ch = getch();
				nctx.linebuf[nctx.linepos++] = ch;
				nctx.wlinebuf[nctx.wlinepos++][1] = ch;
				addch(ch);

				refresh();
				continue;
			}

			/* we have a trailing '\n', now insert '\0' */
			nctx.linebuf[nctx.linepos++] = '\0';
			nctx.wlinebuf[nctx.wlinepos++][0] = '\0';
			clear_input_line(1);
			print_prompt(pctx.nick);

			len = strlen(nctx.linebuf);

			/* chat */
			if (len > 1 && nctx.linebuf[0] != '/') {
				/* delete trailing '\n' */
				nctx.linebuf[len-1] = '\0';
				len--;

				pc_engine_cmd_send_chat(&pctx, nctx.linebuf, len+1);
			}

			/* command */
			if (len > 2 && nctx.linebuf[0] == '/') {
				switch (nctx.linebuf[1]) {
				case 'h':
					pc_engine_cmd_print_help();
					break;
				case 'i':
					pc_engine_cmd_print_info(&pctx);
					break;
				case 'm':
					pc_engine_cmd_toggle_mute(&pctx);
					break;
				case 'l':
					pc_engine_cmd_toggle_loopback(&pctx);
					break;
				case 'r':
				case 'R':
					ret = pc_engine_cmd_toggle_recording(&pctx, nctx.linebuf[1] == 'r' ? REC_OPUS : REC_WAVE);
					if (ret == -1)
						msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "could not start recording");
					break;
				case 'g':
					if (len > 4) {
						float gain = 1.0f;
						sscanf(nctx.linebuf+3, "%f", &gain);
						ret = pc_engine_cmd_set_micgain(&pctx, gain);
						if (ret == -1)
							msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "could not set micgain");
					}
					break;
				case 'G':
					if (len > 4) {
						int idx = -1;
						float gain_dB = 0.0f;
						sscanf(nctx.linebuf+3, "%d %f", &idx, &gain_dB);
						ret = pc_engine_cmd_set_nodegain(&pctx, idx, gain_dB);
						if (ret == -1)
							msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "could not set node gain");
					}
					break;
				case 'b':
					if (len > 4) {
						int bitrate = 16000;
						sscanf(nctx.linebuf+3, "%d", &bitrate);
						ret = pc_engine_cmd_set_bitrate(&pctx, bitrate);
						if (ret == -1)
							msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "could not set bitrate");
					}
					break;
				case 'k':
					if (len > 4) {
						int idx = -1;
						idx = atoi(nctx.linebuf+3);
						ret = pc_engine_cmd_kill_node(&pctx, idx);
						if (ret == -1)
							msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, "could not kill node");
					}
					break;
				case 'c':
					if (len > 4) {
						char *pc;
						int port = 8110;

						/* delete trailing '\n' */
						nctx.linebuf[len-1] = '\0';

						/* extract port */
						pc = strchr(nctx.linebuf+3, ':');
						if (pc) {
							*pc = '\0';
							pc++;
							if (*pc)
								port = atoi(pc);
						}

						/* extract address */
						pc = nctx.linebuf+3;
						while (*pc == ' ')
							pc++;

						ret = pc_engine_cmd_call_node(&pctx, pc, port);
						if (ret == -1) {
							snprintf(msgbuf, MBS, "could not call '%s:%d'", pc, port);
							msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
						}
					}
					break;
				case 'e':
					pc_engine_cmd_toggle_encryption(&pctx);
					break;
				case 'x':
					if (len > 4) {
						char *pc, buf[64];
						size_t klen;

						/* delete trailing '\n' */
						nctx.linebuf[len-1] = '\0';

						/* skip blank spaces */
						pc = nctx.linebuf+3;
						while (*pc == ' ')
							pc++;

						strncpy(buf, pc, 32);
						buf[32] = '\0';
						klen = strlen(buf);
						if (klen > 0) {
							/* key must be 128bit = 16 bytes = 32 hex characters.
							   if we have less characters, fill with '0' */
							if (klen < 32) {
								size_t l;
								for (l = klen; l < 32; l++)
									buf[l] = '0';
							}
							pc_engine_cmd_set_key(&pctx, buf);
						}
					}
					break;
				case 'q':
					keep_going = 0;
					break;
				default:
					/* delete trailing '\n' */
					nctx.linebuf[len-1] = '\0';

					snprintf(msgbuf, MBS, "Unknown command '%s'", nctx.linebuf);
					msgbook_enqueue(&mb0, MB_TYPE_ERROR, MODULE, msgbuf);
					break;
				}
				handle_events(&pctx);
			}
		}


		/** AUDIO CAPTURE **/
		pc_engine_fix_revents(&pctx, pfds);

		if (pfds[fd_idx_audiocapture].revents & POLLERR) {
			if (!(pfds[fd_idx_audiocapture].revents & POLLIN))
				die("Audio Capture POLLERR", 1);
		}

		/* process captured audio if ready */
		if (pfds[fd_idx_audiocapture].revents & POLLIN && keep_going_capture) {

			pc_engine_step(&pctx);
			pc_engine_audio_ready(&pctx);
			handle_events(&pctx);

			/* time to exit? */
			if (keep_going == 0) {
				keep_going_capture = 0;
				pc_engine_goodbye(&pctx);
			}

			/* status bar */
			if (pctx.ic_total % 10 == 0) {
				int y, x, peakcolor, n, h;

				getyx(stdscr, y, x);

				if (pctx.micmute)
					peakcolor = 13;
				else
					peakcolor = (pctx.peak_percent <= 50.0f) ? 11 : ((pctx.peak_percent <= 80.0f) ? 12 : 13);

				move(nctx.rows - 2, 0);
				attron(COLOR_PAIR(10));
				printw("\r%02u:%02u:%02u[", (pctx.ic_talk+pctx.ic_mute)/180000, ((pctx.ic_talk+pctx.ic_mute)/3000)%60,
				       ((pctx.ic_talk+pctx.ic_mute)/50)%60);
				if (pctx.record) {
					attron(COLOR_PAIR(13));
					printw("REC");
					attron(COLOR_PAIR(10));
				} else {
					printw("   ");
				}
				printw("|");
				if (pctx.encrypt) {
					attron(COLOR_PAIR(13));
					printw("ENC");
					attron(COLOR_PAIR(10));
				} else {
					printw("   ");
				}

				printw("]  mic[");
				n = pctx.micmute ? -90 : lrint(pctx.dBSPL);
				n = (n < -50) ? 0 : (n+50); /* min -50dB */
				n = (n + 2) / 3;            /* 3dB step  */
				/*printw("%+05.1f|", pctx.dBSPL);*/
				for (h = 0; h < 14; h++)
					printw("%c", h < n ? '#' : ' ');
				printw("|");
				attron(COLOR_PAIR(peakcolor));
				if (pctx.micmute)
					printw("MUTE");
				else
					printw("%3.0f%%", pctx.peak_percent);
				attron(COLOR_PAIR(10));

				printw("]  %4.1fkB/s↑  %4.1fkB/s↓   pl[%4.2f%%]", pctx.bw_upload, pctx.bw_download,
				       pctx.packets_received_g ?
				       100.0f * pctx.packets_lost_g / (pctx.packets_lost_g + pctx.packets_received_g) : 0.0f);
				attroff(COLOR_PAIR(10));
				chgat(-1, A_NORMAL, 10, NULL);
				move(y, x);
				refresh();

				pctx.sos   = 0;
				pctx.sos_N = 0;
				pctx.peak_percent = 0.0f;
			}
		}
	}

	/* check if we still have stuff on the socket */
	while ((pollret = poll(pfds+fd_idx_udpsocket, 1, 250)) > 0) {
		if (pfds[fd_idx_udpsocket].revents & POLLIN)
			receive_udp_packet(pctx.sockfd, &udp_packet);
	}

	/* clean up */
	pc_engine_cleanup(&pctx);
	handle_events(&pctx);
	free(options.audiodevice_playback);
	free(options.audiodevice_capture);

	/* exit */
	endwin();
	return 0;
}

#undef MODULE
